__all__ = ['Reusable']

import asyncio
import weakref
from dataclasses import dataclass, field
from functools import partial
from threading import Thread

from ._cache import memoize
from ._types import Callback, Get


@memoize()
def make_loop() -> asyncio.AbstractEventLoop:
    loop = asyncio.new_event_loop()
    Thread(target=loop.run_forever, daemon=True).start()
    return loop


async def _await[T](fn: Get[T]) -> T:
    return fn()


def _trampoline[T](callback: Callback[T], ref: weakref.ref[T]) -> None:
    if (obj := ref()) is not None:
        callback(obj)


@dataclass
class Reusable[T]:
    make: Get[T]
    delay: float
    finalize: Callback[T] | None = None

    _loop: asyncio.AbstractEventLoop = field(default_factory=make_loop)
    _deleter: asyncio.TimerHandle | None = None
    _box: list[T] = field(default_factory=list)
    _lock: asyncio.Lock = field(default_factory=asyncio.Lock)

    def __call__(self) -> T:
        """Retrieve or recreate inner object."""
        fut = asyncio.run_coroutine_threadsafe(self._get(), self._loop)
        return fut.result()

    async def _get(self) -> T:
        async with self._lock:
            # reschedule finalizer
            if self._deleter is not None:
                self._deleter.cancel()
            self._deleter = self._loop.call_later(self.delay, self._box.clear)

            # recreate object, if nessessary
            if not self._box:
                obj = await asyncio.to_thread(self.make)
                if self.finalize is not None:
                    weakref.ref(obj, partial(_trampoline, self.finalize))
                self._box.append(obj)

            return self._box[0]



from collections.abc import Callable
from contextlib import AbstractContextManager
from typing import overload

from ._types import Callback, Decorator, Get

def memprof(
    name_or_callback: str | Callback[float] | None = ..., /
) -> AbstractContextManager[None]: ...
@overload
def timer(
    name: str | None = ...,
    time: Get[int] = ...,
    /,
    *,
    disable: bool = ...,
) -> AbstractContextManager[None]: ...
@overload
def timer(
    callback: Callback[int] | None,
    time: Get[int] = ...,
    /,
    *,
    disable: bool = ...,
) -> AbstractContextManager[None]: ...
@overload
def time_this[**P, R](
    fn: Callable[P, R], /, *, name: str | None = ..., disable: bool = ...
) -> Callable[P, R]: ...
@overload
def time_this(*, name: str | None = ..., disable: bool = ...) -> Decorator: ...



__all__ = [
    'call_once',
    'shared_call',
    'streaming',
    'threadlocal',
    'weak_memoize',
]

import sys
import threading
from collections.abc import Callable, Sequence
from concurrent.futures import Future, wait
from functools import partial, update_wrapper
from queue import Empty, SimpleQueue
from threading import Lock, Thread
from time import monotonic, sleep
from typing import Never, cast
from warnings import warn

from ._cache import memoize
from ._dev import hide_frame
from ._futures import BatchDecorator, BatchFn, Job, dispatch, gather_fs
from ._types import Get

_PATIENCE = 0.01


def threadlocal[**P, T](
    fn: Callable[P, T], /, *args: P.args, **kwargs: P.kwargs
) -> Get[T]:
    """Create thread-local singleton factory function (functools.partial)."""
    local_ = threading.local()

    def wrapper() -> T:
        try:
            return local_.obj
        except AttributeError:
            local_.obj = fn(*args, **kwargs)
            return local_.obj

    return update_wrapper(wrapper, fn)


def call_once[T](fn: Get[T], /) -> Get[T]:
    """Make callable a singleton.

    Supports async-def functions (but not async-gen functions).
    DO NOT USE with recursive functions
    """
    warn(
        'Deprecated. Use `@memoize()` for this',
        DeprecationWarning,
        stacklevel=2,
    )
    return memoize()(fn)


def shared_call[**P, R](fn: Callable[P, R], /) -> Callable[P, R]:
    """Merge duplicate parallel invocations of callable to a single one.

    Supports async-def functions (but not async-gen functions).
    DO NOT USE with recursive functions
    """
    warn(
        'Deprecated. Use `@memoize(0)` for this',
        DeprecationWarning,
        stacklevel=2,
    )
    return memoize(0)(fn)


def weak_memoize[**P, R](fn: Callable[P, R], /) -> Callable[P, R]:
    """Preserve each result of each call until they are garbage collected."""
    warn(
        'Deprecated. Use `@memoize(0)` for this',
        DeprecationWarning,
        stacklevel=2,
    )
    return memoize(0)(fn)


# ----------------------------- batch collation ------------------------------


def _fetch_batch[T](
    q: SimpleQueue[T], batch_size: int | None, timeout: float
) -> list[T]:
    batch: list[T] = []

    # Wait indefinitely until the first item is received
    if sys.platform == 'win32':
        # On Windows lock.acquire called without a timeout is not interruptible
        # See:
        # https://bugs.python.org/issue29971
        # https://github.com/dask/dask/pull/2144#issuecomment-290556996
        # https://github.com/dask/dask/pull/2144/files
        while not batch:
            try:
                batch.append(q.get(timeout=_PATIENCE))
            except Empty:
                sleep(0)  # Allow other thread to fill the batch
    else:
        batch.append(q.get())

    endtime = monotonic() + timeout
    remaining = timeout
    while remaining > 0 and (batch_size is None or len(batch) < batch_size):
        try:
            batch.append(q.get(timeout=remaining))
        except Empty:
            break
        remaining = endtime - monotonic()
    return batch


def _start_fetch_compute[T, R](
    func: BatchFn[T, R],
    workers: int,
    batch_size: int | None,
    timeout: float,
) -> SimpleQueue[Job[T, R]]:
    q = SimpleQueue()  # type: ignore[var-annotated]
    lock = Lock()

    def loop() -> Never:
        while True:
            # Because of lock, _fetch_batch could be inlined into wrapper,
            # and dispatch to thread pool could be done from there,
            # thus allowing usage of scalable ThreadPool
            # TODO: implement above
            with lock:  # Ensurance that none worker steals tasks from other
                batch = _fetch_batch(q, batch_size, timeout)
            if batch := [
                (x, f) for x, f in batch if f.set_running_or_notify_cancel()
            ]:
                dispatch(func, *batch)
            else:
                sleep(0.001)

    for _ in range(workers):
        Thread(target=loop, daemon=True).start()
    return q


def streaming[T, R](
    func: BatchFn[T, R] | None = None,
    /,
    *,
    batch_size: int | None = None,
    timeout: float = 0.1,
    workers: int = 1,
    pool_timeout: float = 20.0,
) -> BatchDecorator | BatchFn[T, R]:
    """Delay start of computation to until batch is collected.

    Accepts two timeouts (in seconds):
    - `timeout` is a time to wait till the batch is full, i.e. latency.
    - `pool_timeout` is time to wait for results.

    Uses ideas from
    - https://github.com/ShannonAI/service-streamer
    - https://github.com/leon0707/batch_processor
    - ray.serve.batch
      https://github.com/ray-project/ray/blob/master/python/ray/serve/batching.py

    Note: currently supports only functions and bound methods.

    Implementation details:
    - constantly keeps alive N workers
    - any caller enqueues jobs and starts waiting
    - on any failure during waiting caller cancels all jobs it submitted
    - single worker at a time fetches jobs from shared queue, resolves them,
      and notifies all waiters
    """
    if func is None:
        deco = partial(
            streaming,
            batch_size=batch_size,
            timeout=timeout,
            workers=workers,
            pool_timeout=pool_timeout,
        )
        return cast('BatchDecorator', deco)

    assert callable(func)
    assert workers >= 1
    q = _start_fetch_compute(func, workers, batch_size, timeout)

    def wrapper(items: Sequence[T]) -> Sequence[R]:
        fs = {Future[R](): item for item in items}
        try:
            for f, x in fs.items():
                q.put((x, f))  # Schedule task
            dnd = wait(fs, pool_timeout, return_when='FIRST_EXCEPTION')

        finally:  # Cancel all not-yet-running tasks, we're beyond deadline
            for f in fs:
                f.cancel()

        if dnd.not_done:  # Some tasks timed out
            del dnd, fs  # ? Break reference cycle
            raise TimeoutError

        # Cannot time out - all are done
        rs, err = gather_fs(enumerate(fs))
        if err is None:
            return list(rs.values())
        with hide_frame:
            raise err

    # TODO: if func is instance method - recreate wrapper per instance
    # TODO: find how to distinguish between
    # TODO:  not yet bound method and plain function
    # TODO:  maybe implement __get__ on wrapper
    return update_wrapper(wrapper, func)



__all__ = [
    'buffered',
    'get_executor',
    'map_n',
    'map_n_dict',
    'max_cpu_count',
    'starmap_n',
]

import atexit
import enum
import logging
import os
import signal
import sys
import warnings
import weakref
from collections.abc import Callable, Iterable, Iterator, Mapping, Sized
from concurrent.futures import Executor, Future
from contextlib import ExitStack, contextmanager
from cProfile import Profile
from functools import partial
from itertools import chain, islice, starmap
from multiprocessing.managers import BaseManager
from operator import methodcaller
from pstats import Stats
from queue import Empty, SimpleQueue
from threading import Lock
from time import monotonic, sleep
from typing import Final, Protocol, Self, cast

import loky

try:
    import psutil
except ImportError:
    psutil = None

from ._dev import hide_frame
from ._more import chunked, ilen
from ._reduction import move_to_shmem, reducers
from ._thread_quota import ThreadQuota
from ._types import Get, Some

_LOGGER = logging.getLogger(__name__)

_TOTAL_CPUS = (
    os.process_cpu_count() if sys.version_info >= (3, 13) else os.cpu_count()
)
_NUM_CPUS = _TOTAL_CPUS or 0

if (_env_cpus := os.getenv('GLOW_CPUS')) is not None:
    _NUM_CPUS = min(_NUM_CPUS, int(_env_cpus))
    _NUM_CPUS = max(_NUM_CPUS, 0)

_IDLE_WORKER_TIMEOUT = 10
# TODO: investigate whether this improves load
_FAST_GROW = False
_GRANULAR_SCHEDULING = False


class _Empty(enum.Enum):
    token = 0


_empty: Final = _Empty.token

# ------------------- some useful interfaces and functions -------------------


class _Queue[T](Protocol):
    def get(self, block: bool = ..., timeout: float | None = ...) -> T: ...
    def put(self, item: T) -> None: ...


class _Event(Protocol):
    def is_set(self) -> bool: ...
    def set(self) -> None: ...


class _Manager(Protocol):
    def Event(self) -> _Event: ...  # noqa: N802
    def Queue(self, /, maxsize: int) -> _Queue: ...  # noqa: N802


def _torch_limit() -> int | None:
    # Windows platform lacks memory overcommit, so it's sensitive to VMS growth
    if sys.platform != 'win32':
        return None

    torch = sys.modules.get('torch')
    if torch is None or (torch.version.cuda or '') >= '11.7.0':
        # It's expected that torch will fix .nv_fatb readonly flag in its DLLs
        # See https://stackoverflow.com/a/69489193/9868257
        return None

    if psutil is None:
        warnings.warn(
            'Max process count may be calculated incorrectly, '
            'leading to application crash or even BSOD. '
            'Install psutil to avoid that',
            stacklevel=3,
        )
        return None

    # Windows has no overcommit, checking how much processes fit into VMS
    vms: int = psutil.Process().memory_info().vms
    free_vms: int = psutil.virtual_memory().free + psutil.swap_memory().free
    return free_vms // vms


def max_cpu_count(limit: int | None = None, *, mp: bool = False) -> int:
    limits = [_TOTAL_CPUS or 1]

    if limit is not None:
        limits.append(limit)

    if mp and (torch_limit := _torch_limit()) is not None:
        limits.append(torch_limit)

    return min(limits)


_PATIENCE = 0.01


class _TimeoutCallable[T](Protocol):
    def __call__(self, *, timeout: float) -> T: ...


def _retrying[T](f: _TimeoutCallable[T], *exc: type[BaseException]) -> T:
    # See issues
    # https://github.com/dask/dask/pull/2144#issuecomment-290556996
    # https://github.com/dask/dask/pull/2144/files
    # https://github.com/python/cpython/issues/74157
    # FIXED in py3.15+
    while True:
        try:
            return f(timeout=_PATIENCE)
        except exc:
            sleep(0)  # Force switch to another thread to proceed


if sys.platform == 'win32':

    def _exception[T](f: Future[T], /) -> BaseException | None:
        return _retrying(f.exception, TimeoutError)

else:
    _exception = Future.exception


def _result[T](f: Future[T], cancel: bool = True) -> Some[T] | BaseException:
    try:
        return exc if (exc := _exception(f)) else Some(f.result())
    finally:
        if cancel:
            f.cancel()
        del f


def _q_get_fn[T](q: _Queue[T]) -> Get[T]:
    if sys.platform != 'win32':
        return q.get
    return partial(_retrying, q.get, Empty)


# ---------------------------- pool initialization ----------------------------


def _mp_profile() -> None:
    """Multiprocessed profiler."""
    prof = Profile()
    prof.enable()

    def _finalize(lines=50) -> None:
        prof.disable()
        with open(f'prof-{os.getpid()}.txt', 'w') as fp:
            Stats(prof, stream=fp).sort_stats('cumulative').print_stats(lines)

    atexit.register(_finalize)


def _initializer() -> None:
    # `signal.signal` suppresses KeyboardInterrupt in child processes
    signal.signal(signal.SIGINT, signal.SIG_IGN)
    if os.environ.get('_GLOW_MP_PROFILE'):
        _mp_profile()


@contextmanager
def get_executor(max_workers: int, *, mp: bool) -> Iterator[Executor]:
    if mp:
        processes: loky.ProcessPoolExecutor = loky.get_reusable_executor(
            max_workers,
            'loky_init_main',
            _IDLE_WORKER_TIMEOUT,
            job_reducers=reducers,
            result_reducers=reducers,
            initializer=_initializer,
        )
        # In generator 'finally' is not reliable enough, use atexit
        hook = atexit.register(processes.shutdown, kill_workers=True)
        yield processes
        atexit.unregister(hook)
    else:
        threads = ThreadQuota(max_workers)
        try:
            yield threads
        finally:
            is_success = sys.exception() is None
            threads.shutdown(wait=is_success, cancel_futures=True)


def _get_manager(executor: Executor) -> _Manager:
    if isinstance(executor, loky.ProcessPoolExecutor):
        return executor._context.Manager()

    else:  # noqa: RET505
        from multiprocessing.dummy import Manager

        return Manager()


# -------- bufferize iterable by offloading to another thread/process --------


def _consume[T](
    items: Iterable[T], buf: _Queue[T | _Empty], stop: _Event
) -> None:
    try:
        for item in items:
            if stop.is_set():
                break
            buf.put(item)
    finally:
        buf.put(_empty)  # Signal to stop iteration
        buf.put(_empty)  # Match last q.get


class buffered[T](Iterator[T]):  # noqa: N801
    """Iterate in background thread with at most `latency` items ahead."""

    __slots__ = ('__weakref__', '_consume', '_next', 'close')

    def __init__(
        self,
        iterable: Iterable[T],
        /,
        *,
        latency: int = 2,
        mp: bool | Executor = False,
    ) -> None:
        s = ExitStack()
        if isinstance(mp, Executor):
            executor = mp
        else:
            executor = s.enter_context(get_executor(1, mp=mp))

        mgr = _get_manager(executor)
        if isinstance(mgr, BaseManager):
            s.enter_context(mgr)

        ev: _Event = mgr.Event()
        q: _Queue[T | _Empty] = mgr.Queue(latency)
        self._consume = executor.submit(_consume, iterable, q, ev)
        self._next = _q_get_fn(q)

        # If main is killed, unblocks consumer to allow it to check stop flag
        # Otherwise collects 2nd _empty from q.
        # Called 2nd
        s.callback(self._next)

        # If main is killed, notifies consumer to stop.
        # If consumer is already stopped (on error or not), does nothing.
        # Called 1st
        s.callback(ev.set)

        self.close = weakref.finalize(self, s.close)

    def __iter__(self) -> Self:
        return self

    def __next__(self) -> T:
        if self.close.alive:
            if (item := self._next()) is not _empty:
                return item

            self.close()
            # Reraise exception from source iterable if any
            obj = _result(self._consume, cancel=False)
            if not isinstance(obj, Some):
                with hide_frame:
                    raise obj

        raise StopIteration


# ---------------------------- automatic batching ----------------------------


class _AutoSize:
    MIN_DURATION = 0.2
    MAX_DURATION = 2.0
    size: int = 1
    duration: float = 0.0

    def __init__(self) -> None:
        self.lock = Lock()
        assert self.MIN_DURATION * 2 < self.MAX_DURATION, 'Range is too tight'

    def suggest(self) -> int:
        with self.lock:
            return self.size

    def update(self, n: int, start_time: float, fut: Future[Sized]) -> None:
        # Compute as soon as future became done, discard later if not needed
        duration = monotonic() - start_time

        if fut.cancelled():  # Job never run, zero load
            return

        with self.lock:
            if n != self.size:  # Ran with old size
                return

            # Do EMA smoothing
            self.duration = (
                (0.8 * self.duration + 0.2 * duration)
                if self.duration > 0
                else duration
            )
            if self.duration <= 0:  # Smh not initialized yet
                return  # Or duration is less then `monotonic()` precision

            if self.duration < self.MIN_DURATION:  # Too high IPC overhead
                size = self._new_scale() if _FAST_GROW else self.size * 2
                _LOGGER.debug('Increasing batch size to %d', size)

            elif (
                self.duration <= self.MAX_DURATION  # Range is optimal
                or self.size == 1  # Cannot reduce already minimal batch
            ):
                return

            else:  # Too high latency
                size = self._new_scale()
                _LOGGER.debug('Reducing batch size to %d', size)

            self.size = size
            self.duration = 0.0

    def _new_scale(self) -> int:
        factor = 2 * self.MIN_DURATION / self.duration
        factor = min(factor, 32)
        size = int(self.size * factor)
        return max(size, 1)


# ---------------------- map iterable through function ----------------------


def _schedule[F: Future](
    submit_chunk: Callable[..., F],
    args_zip: Iterable[Iterable],
    chunksize: int,
) -> Iterator[F]:
    for chunk in chunked(args_zip, chunksize):
        f = submit_chunk(*chunk)
        _LOGGER.debug('Submit %d', len(chunk))
        yield f


def _schedule_auto[F: Future](
    submit_chunk: Callable[..., F],
    args_zip: Iterator[Iterable],
    max_workers: int,
) -> Iterator[F]:
    # For the whole wave make futures with the same job size
    size = _AutoSize()
    while tuples := [*islice(args_zip, size.suggest() * max_workers)]:
        chunksize = len(tuples) // max_workers or 1
        for chunk in chunked(tuples, chunksize):
            f = submit_chunk(*chunk)
            _LOGGER.debug('Submit %d', len(chunk))
            f.add_done_callback(partial(size.update, len(chunk), monotonic()))
            yield f


def _schedule_auto_v2[F: Future](
    submit_chunk: Callable[..., F], args_zip: Iterator[Iterable]
) -> Iterator[F]:
    # Vary job size from future to future
    size = _AutoSize()
    while chunk := [*islice(args_zip, size.suggest())]:
        f = submit_chunk(*chunk)
        _LOGGER.debug('Submit %d', len(chunk))
        f.add_done_callback(partial(size.update, len(chunk), monotonic()))
        yield f


def _get_unwrap_iter[T](
    s: ExitStack,
    qsize: int,
    get_f: Get[Future[T]],
    sched_iter: Iterator,
) -> Iterator[T]:
    with s:
        if not qsize:  # No tasks to do
            return

        # Unwrap 1st / schedule `N-qsize` / unwrap `qsize-1`
        with hide_frame:
            for _ in chain([None], sched_iter, range(qsize - 1)):
                # Retrieve done task, exactly `N` calls
                obj = _result(get_f())
                if not isinstance(obj, Some):
                    with hide_frame:
                        raise obj
                yield obj.x


def _enqueue[T](
    fs: Iterable[Future[T]],
    unordered: bool,
) -> tuple[Iterator, Get[Future[T]]]:
    q = SimpleQueue[Future[T]]()

    # In `unordered` mode `q` contains only "DONE" tasks,
    # else there are also "PENDING" and "RUNNING" tasks.
    # FIXME: unordered=True -> random freezes (in q.get -> Empty)
    q_put = cast(
        'Callable[[Future[T]], None]',
        methodcaller('add_done_callback', q.put) if unordered else q.put,
    )
    q_get = _q_get_fn(q)

    # On each `next()` schedules new task
    sched_iter = map(q_put, fs)

    return sched_iter, q_get


def _prefetch(s: ExitStack, sched_iter: Iterator, count: int | None) -> int:
    try:
        # Fetch up to `count` tasks to pre-fill `q`
        qsize = ilen(islice(sched_iter, count))
    except BaseException:
        # Unwind stack here on an error
        s.close()
        raise
    else:
        _LOGGER.debug('Prefetched %d jobs', qsize)
        return qsize


def _batch_invoke[*Ts, R](
    func: Callable[[*Ts], R], *items: tuple[*Ts]
) -> list[R]:
    return [*starmap(func, items)]


def starmap_n[T](
    func: Callable[..., T],
    iterable: Iterable[Iterable],
    /,
    *,
    max_workers: int | None = None,
    prefetch: int | None = 2,
    mp: bool = False,
    chunksize: int | None = None,
    unordered: bool = False,
) -> Iterator[T]:
    """Equivalent to itertools.starmap(fn, iterable).

    Return an iterator whose values are returned from the function evaluated
    with an argument tuple taken from the given sequence.

    Options:
    - workers - Count of workers, by default all hardware threads are occupied.
    - prefetch - Count of extra jobs to schedule over N workers.
      Helps with CPU stalls in ordered mode.
      Increase if job execution time is highly variable.
    - mp - Whether use processes or threads.
    - chunksize - The size of the chunks the iterable will be broken into
      before being passed to a processes.
      Estimated automatically.
      Ignored when threads are used.
    - unordered - Retrieve results in order of completion or in original order.
      In this mode `prefetch` is meaningless, because when some job became done
      it yielded immediately releasing buffer for new job to schedule.
      So no CPU stalls.

    Unlike multiprocessing.Pool or concurrent.futures.Executor this one:
    - never deadlocks on any exception or Ctrl-C interruption.
    - accepts infinite iterables due to lazy task creation.
    - has single interface for both threads and processes.
    - TODO: serializes array-like data using out-of-band Pickle 5 buffers.
    - call immediately creates pool ready to yield results
      (which could take some time cause of serialization for multiprocessing),
      so first `__next__` runs on warmed up pool.
    """
    if max_workers is None:
        max_workers = max_cpu_count(_NUM_CPUS, mp=mp)

    if not max_workers or not _NUM_CPUS:
        return starmap(func, iterable)  # Fallback to single thread

    if mp and chunksize is None and prefetch is None:
        msg = 'With multiprocessing either chunksize or prefetch should be set'
        raise ValueError(msg)

    if unordered:
        prefetch = max(max_workers, 1)
    elif prefetch is not None:
        prefetch = max(prefetch + max_workers, 1)

    it = iter(iterable)
    s = ExitStack()
    submit = s.enter_context(get_executor(max_workers, mp=mp)).submit

    if mp:
        func = move_to_shmem(func)
    else:
        chunksize = chunksize or 1

    if chunksize == 1:
        submit_1 = cast('Callable[..., Future[T]]', partial(submit, func))
        f1s = starmap(submit_1, it)
        sched1_iter, get_f = _enqueue(f1s, unordered)
        qsize = _prefetch(s, sched1_iter, prefetch)
        return _get_unwrap_iter(s, qsize, get_f, sched1_iter)

    submit_n = cast(
        'Callable[..., Future[list[T]]]', partial(submit, _batch_invoke, func)
    )
    if chunksize is not None:
        # Fixed chunksize
        fs = _schedule(submit_n, it, chunksize)
    elif not _GRANULAR_SCHEDULING:
        # Dynamic chunksize scaling, submit tasks in waves
        fs = _schedule_auto(submit_n, it, max_workers)
    else:
        # Dynamic chunksize scaling
        fs = _schedule_auto_v2(submit_n, it)

    sched_iter, get_fs = _enqueue(fs, unordered)
    qsize = _prefetch(s, sched_iter, prefetch)
    chunks = _get_unwrap_iter(s, qsize, get_fs, sched_iter)
    return chain.from_iterable(chunks)


def map_n[T](
    func: Callable[..., T],
    /,
    *iterables: Iterable,
    max_workers: int | None = None,
    prefetch: int | None = 2,
    mp: bool = False,
    chunksize: int | None = None,
    unordered: bool = False,
) -> Iterator[T]:
    """Return iterator equivalent to map(func, *iterables).

    Make an iterator that computes the function using arguments from
    each of the iterables. Stops when the shortest iterable is exhausted.

    For extra options, see starmap_n, which is used under hood.
    """
    return starmap_n(
        func,
        zip(*iterables),
        max_workers=max_workers,
        prefetch=prefetch,
        mp=mp,
        chunksize=chunksize,
        unordered=unordered,
    )


def map_n_dict[K, T1, T2](
    func: Callable[[T1], T2],
    obj: Mapping[K, T1],
    /,
    *,
    max_workers: int | None = None,
    prefetch: int | None = 2,
    mp: bool = False,
    chunksize: int | None = None,
) -> dict[K, T2]:
    """Apply `func` to each value in a mapping in parallel way.

    For extra options, see starmap_n, which is used under hood.
    """
    iter_values = map_n(
        func,
        obj.values(),
        max_workers=max_workers,
        prefetch=prefetch,
        mp=mp,
        chunksize=chunksize,
    )
    return dict(zip(obj.keys(), iter_values, strict=True))



__all__ = ['clone_exc', 'hide_frame']

from types import TracebackType


class _HideFrame:
    """Context manager to hide current frame in traceback"""

    def __enter__(self):
        return self

    def __exit__(
        self, tp, val: BaseException | None, tb: TracebackType | None
    ):
        if val is not None:
            tb = val.__traceback__ or tb
            if tb:
                val.__traceback__ = tb.tb_next  # Drop outer traceback frame


def clone_exc[E: BaseException](exc: E) -> E:
    new_exc = type(exc)(*exc.args)
    new_exc.__cause__ = exc.__cause__
    new_exc.__context__ = exc.__context__
    new_exc.__traceback__ = exc.__traceback__
    return new_exc


hide_frame = _HideFrame()



# flake8: noqa
"""Functional Python tools"""

from importlib import import_module
from typing import TYPE_CHECKING

from . import _patch_len, _patch_print, _patch_scipy
from ._array import aceil, afloor, apack, around, pascal
from ._async import RwLock, amap, amap_dict, astarmap, astreaming, azip
from ._cache import cache_status, memoize
from ._concurrency import (
    call_once,
    shared_call,
    streaming,
    threadlocal,
    weak_memoize,
)
from ._coro import as_actor, coroutine, summary
from ._debug import lock_seed, trace, trace_module, whereami
from ._dev import hide_frame
from ._import_hook import register_post_import_hook, when_imported
from ._logging import init_loguru
from ._more import (
    as_iter,
    chunked,
    eat,
    groupby,
    ichunked,
    ilen,
    roundrobin,
    windowed,
)
from ._parallel import (
    buffered,
    get_executor,
    map_n,
    map_n_dict,
    max_cpu_count,
    starmap_n,
)
from ._profile import memprof, time_this, timer
from ._repr import countable, mangle, repr_as_obj, si, si_bin
from ._reusable import Reusable
from ._sizeof import sizeof
from ._streams import cumsum, maximum_cumsum
from ._uuid import Uid

if TYPE_CHECKING:
    from ._ic import ic, ic_repr
    from ._imutil import imhash_hist
else:
    _exports = {
        '._ic': ['ic', 'ic_repr'],
        '._imutil': ['imhash_hist'],
    }
    _submodule_by_name = {
        name: modname for modname, names in _exports.items() for name in names
    }

    def __getattr__(name: str):
        if mod := _submodule_by_name.get(name):
            mod = import_module(mod, __package__)
            globals()[name] = obj = getattr(mod, name)
            return obj
        raise AttributeError(f'No attribute {name}')

    def __dir__():
        return __all__


__all__ = [
    'Reusable',
    'RwLock',
    'Uid',
    'aceil',
    'afloor',
    'amap',
    'amap_dict',
    'apack',
    'around',
    'as_actor',
    'as_iter',
    'astarmap',
    'astreaming',
    'azip',
    'buffered',
    'cache_status',
    'call_once',
    'chunked',
    'coroutine',
    'countable',
    'cumsum',
    'eat',
    'get_executor',
    'groupby',
    'hide_frame',
    'ic',
    'ic_repr',
    'ichunked',
    'ilen',
    'imhash_hist',
    'init_loguru',
    'lock_seed',
    'mangle',
    'map_n',
    'map_n_dict',
    'max_cpu_count',
    'maximum_cumsum',
    'memoize',
    'memprof',
    'pascal',
    'register_post_import_hook',
    'repr_as_obj',
    'roundrobin',
    'shared_call',
    'si',
    'si_bin',
    'sizeof',
    'starmap_n',
    'streaming',
    'summary',
    'threadlocal',
    'time_this',
    'timer',
    'trace',
    'trace_module',
    'weak_memoize',
    'when_imported',
    'whereami',
    'windowed',
]

_patch_print.apply()
_patch_len.apply()
_patch_scipy.apply()



__all__ = ['memprof', 'time_this', 'timer']

import atexit
from collections import defaultdict
from collections.abc import Callable, Iterator
from contextlib import AbstractContextManager, contextmanager, nullcontext
from dataclasses import dataclass, field
from functools import partial
from itertools import count
from time import perf_counter_ns, process_time_ns, thread_time_ns
from typing import TYPE_CHECKING

from ._debug import whereami
from ._repr import si, si_bin
from ._streams import Stream, cumsum, maximum_cumsum
from ._types import Callback, Get

from ._wrap import wrap

if TYPE_CHECKING:
    import psutil

    _THIS: psutil.Process | None

_THIS = None


@contextmanager
def memprof(
    name_or_callback: str | Callback[float] | None = None, /
) -> Iterator[None]:
    global _THIS  # noqa: PLW0603
    if _THIS is None:
        import psutil  # noqa: PLC0415

        _THIS = psutil.Process()

    init = _THIS.memory_info().rss
    try:
        yield
    finally:
        size = _THIS.memory_info().rss - init
        if callable(name_or_callback):
            name_or_callback(size)
        else:
            name = name_or_callback
            if name is None:
                name = f'{whereami(2, 1)} line'
            sign = '+' if size >= 0 else ''
            print(f'{name} done: {sign}{si_bin(size)}')


@contextmanager
def _timer_callback(
    callback: Callback[int], time: Get[int] = perf_counter_ns, /
) -> Iterator[None]:
    begin = time()
    try:
        yield
    finally:
        callback(time() - begin)


@contextmanager
def _timer_print(
    name: str | None = None, time: Get[int] = perf_counter_ns, /
) -> Iterator[None]:
    begin = time()
    try:
        yield
    finally:
        end = time()
        name = name or f'{whereami(2, 1)} line'
        print(f'{name} done in {si((end - begin) / 1e9)}s')


def timer(
    name_or_callback: str | Callback[int] | None = None,
    time: Get[int] = perf_counter_ns,
    /,
    *,
    disable: bool = False,
) -> AbstractContextManager[None]:
    if disable:
        return nullcontext()
    if callable(name_or_callback):
        return _timer_callback(name_or_callback, time)
    return _timer_print(name_or_callback, time)


def _to_fname(obj) -> str:
    if not hasattr(obj, '__module__') or not hasattr(obj, '__qualname__'):
        obj = type(obj)
    if obj.__module__ == 'builtins':
        return obj.__qualname__
    return f'{obj.__module__}.{obj.__qualname__}'


@dataclass(frozen=True, slots=True)
class _Profiler:
    # Call count
    calls: count = field(default_factory=count)
    # Statistics requests count
    reads: count = field(default_factory=count)
    # Parallelism/concurrency
    active_calls: Stream[int, int] = field(default_factory=maximum_cumsum)
    # Time spent executing this thread, including kernel time, but not I/O.
    busy_ns: Stream[int, int] = field(default_factory=cumsum)
    # Idle time - elapsed I/O time (like time.sleep, lock.acquire, e.t.c.).
    idle_ns: Stream[int, int] = field(default_factory=cumsum)

    def suspend(self) -> Get[None]:
        self.idle_ns.send(-perf_counter_ns())
        return self.resume

    def resume(self) -> None:
        self.idle_ns.send(+perf_counter_ns())

    def __call__[**P, R](
        self, op: Callable[P, R], /, *args: P.args, **kwargs: P.kwargs
    ) -> R:
        self.active_calls.send(+1)
        t_cpu = thread_time_ns()
        self.idle_ns.send(+t_cpu - perf_counter_ns())
        self.busy_ns.send(-t_cpu)
        try:
            return op(*args, **kwargs)
        finally:
            t_cpu = thread_time_ns()
            self.busy_ns.send(+t_cpu)
            self.idle_ns.send(-t_cpu + perf_counter_ns())
            self.active_calls.send(-1)

    def stat(self) -> tuple[float, float, str] | None:
        if not (n := next(self.calls) - next(self.reads)):
            return None
        concurrency = self.active_calls.send(0)
        idle = self.idle_ns.send(0) / 1e9  # Wall-clock - user+sys
        busy = self.busy_ns.send(0) / 1e9  # user+system time

        tail = ''
        if n > 1:
            tail += f'{n} x {si(busy / n)}s'
            if concurrency > 1:
                tail += f' @ {concurrency}T'
        return busy, idle, tail


# Wall time, i.e. sum of per-thread times, excluding sleep
_start = process_time_ns()
_profilers = defaultdict[str, _Profiler](_Profiler)


@atexit.register
def _print_stats(*names: str) -> None:
    all_busy = (process_time_ns() - _start + 1) / 1e9

    stats: list[tuple[float, float, str, str]] = []  # (busy, idle, tail, name)
    names = names or tuple(_profilers)
    for name in names:
        if not (profiler := _profilers.pop(name, None)):
            continue
        if not (lines := profiler.stat()):
            continue
        stats.append((*lines, name))

    for busy, idle, tail, name in sorted(stats):
        print(
            f'{busy / all_busy:6.2%}',
            f'{si(busy):>5s}s + {si(idle):>5s}s',
            name,
            tail,
            sep=' - ',
        )


def time_this(fn=None, /, *, name: str | None = None, disable: bool = False):
    """Log function and/or generator timings at program exit."""
    if fn is None:
        return partial(time_this, name=name, disable=disable)
    if disable:
        return fn

    if name is None:
        name = _to_fname(fn)

    fin = fn.log_timing = partial(_print_stats, name)
    time_this.finalizers[fn] = fin  # type: ignore[attr-defined]
    return wrap(fn, _profilers[name])


time_this.finalizers = {}  # type: ignore[attr-defined]



__all__ = ['cache_status', 'memoize']

import asyncio
import concurrent.futures as cf
import enum
import functools
from collections.abc import (
    Awaitable,
    Callable,
    Hashable,
    Iterable,
    Iterator,
    KeysView,
    Mapping,
    MutableMapping,
)
from dataclasses import dataclass, field
from inspect import iscoroutinefunction
from threading import RLock
from typing import Any, Final, Protocol, SupportsInt, cast
from weakref import WeakValueDictionary

from ._dev import clone_exc, hide_frame
from ._futures import (
    ABatchFn,
    AnyFuture,
    BatchFn,
    Job,
    adispatch,
    dispatch,
    gather_fs,
)
from ._keys import make_key
from ._repr import si_bin
from ._sizeof import sizeof
from ._types import CachePolicy, Decorator, KeyFn, Some


class _Empty(enum.Enum):
    token = 0


_empty: Final = _Empty.token


@dataclass(repr=False, slots=True)
class _Node[T]:
    value: T
    size: int

    def __repr__(self) -> str:
        return repr(self.value)


def _make_node[T](obj: T, /) -> _Node[T]:
    return _Node(obj, 1)


def _make_sized_node[T](obj: T, /) -> _Node[T]:
    return _Node(obj, sizeof(obj))


@dataclass
class Stats:
    hits: int = 0
    misses: int = 0
    dropped: int = 0

    def __bool__(self) -> bool:
        return any(self.__dict__.values())

    def __repr__(self) -> str:
        fields = ', '.join(f'{k}={v}' for k, v in self.__dict__.items() if v)
        return f'{self.__class__.__name__}({fields})'


# ----------------------------- basic caches ------------------------------


def cache_status() -> str:
    return '\n'.join(
        f'{id_:x}: {value!r}' for id_, value in sorted(_REFS.items())
    )


_REFS: MutableMapping[int, '_Cache'] = WeakValueDictionary()


class _AbstractCache[K: Hashable, T](Protocol):
    def __getitem__(self, key: K, /) -> T | _Empty: ...
    def __setitem__(self, key: K, value: T, /) -> None: ...


class _CacheMaker[K, T](Protocol):
    def __call__(
        self, capacity: int, make_node: Callable[[T], _Node[T]]
    ) -> '_AbstractCache[K, T]': ...


@dataclass(repr=False, slots=True, weakref_slot=True)
class _Cache[K: Hashable, T]:
    capacity: int
    make_node: Callable[[T], _Node[T]] = field(repr=False)
    size: int = 0
    store: dict[K, _Node[T]] = field(default_factory=dict)
    stats: Stats = field(default_factory=Stats)

    def __post_init__(self) -> None:
        _REFS[id(self)] = self

    def __len__(self) -> int:
        return len(self.store)

    def __iter__(self) -> Iterator:
        return iter(self.store)

    def keys(self) -> KeysView:
        return self.store.keys()

    def clear(self) -> None:
        self.stats.dropped += len(self.store)
        self.store.clear()
        self.size = 0

    def __repr__(self) -> str:
        args = [
            f'items={len(self.store)}',
            f'size={type(self.capacity)(self.size)}',
            f'capacity={self.capacity}',
        ]
        if self.stats:
            args.append(f'stats={self.stats}')
        return f'{type(self).__name__}({", ".join(args)})'


class _Heap[K: Hashable, T](_Cache[K, T]):
    def __getitem__(self, key: K, /) -> T | _Empty:
        if node := self.store.get(key):
            self.stats.hits += 1
            return node.value

        self.stats.misses += 1
        return _empty

    def __setitem__(self, key: K, value: T, /) -> None:
        if key in self.store:
            return
        node = self.make_node(value)
        if (
            self.capacity >= 0  # bound cache
            and self.size + node.size > self.capacity  # no free place
        ):
            return
        self.store[key] = node
        self.size += node.size


class _LruMruCache[K: Hashable, T](_Cache[K, T]):
    def __getitem__(self, key: K, /) -> T | _Empty:
        if node := self.store.pop(key, None):
            self.stats.hits += 1
            self.store[key] = node
            return node.value

        self.stats.misses += 1
        return _empty

    def __setitem__(self, key: K, value: T, /) -> None:
        if key in self.store:
            return
        node = self.make_node(value)
        nsize = node.size
        if self.capacity >= 0:  # bound cache
            if nsize > self.capacity:  # cache will never fit this
                return

            while self.store and self.size + nsize > self.capacity:  # evict
                self.size -= self.pop().size
                self.stats.dropped += 1

        self.store[key] = node
        self.size += nsize

    def pop(self) -> _Node:
        raise NotImplementedError


class _LruCache[K: Hashable, T](_LruMruCache[K, T]):
    def pop(self) -> _Node:
        """Drop oldest node."""
        return self.store.pop(next(iter(self.store)))


class _MruCache[K: Hashable, T](_LruMruCache[K, T]):
    def pop(self) -> _Node:
        """Drop most recently added node."""
        return self.store.popitem()[1]


# --------------------------------- utilities --------------------------------


@dataclass(frozen=True, kw_only=True)
class _WeakCache[K: Hashable, T]:
    """Retrieve items via weak references from everywhere."""

    alive: WeakValueDictionary[K, T] = field(
        default_factory=WeakValueDictionary
    )

    def __getitem__(self, key: K, /) -> T | _Empty:
        return self.alive.get(key, _empty)

    def __setitem__(self, key: K, value: T, /) -> None:
        if type(value).__weakrefoffset__:  # Support weak reference.
            self.alive[key] = value


@dataclass(frozen=True, kw_only=True)
class _StrongCache[K: Hashable, T](_WeakCache[K, T]):
    cache: _AbstractCache[K, T]

    def __getitem__(self, key: K, /) -> T | _Empty:
        # Alive and stored items.
        # Called first to update cache stats (i.e. MRU/LRU if any).
        # `cache` has subset of objects from `alive`.
        if (ret := self.cache[key]) is not _empty:
            return ret
        # Item could still exist, try reference ...
        return super().__getitem__(key)

    def __setitem__(self, key: K, value: T, /) -> None:
        self.cache[key] = value
        super().__setitem__(key, value)


@dataclass(frozen=True, slots=True)
class _CacheState[K: Hashable, R]:
    cache: _AbstractCache[K, R]
    key_fn: KeyFn[K]
    futures: WeakValueDictionary[K, AnyFuture[R]] = field(
        default_factory=WeakValueDictionary
    )


# --------------------------------- wrapping ---------------------------------


def _result[T](f: cf.Future[T]) -> T:
    if f.cancelled():
        with hide_frame:
            raise cf.CancelledError
    if exc := f.exception():
        with hide_frame:
            raise exc
    return f.result()


def _sync_memoize[K: Hashable, **P, R](
    fn: Callable[P, R],
    cs: _CacheState[K, R],
) -> Callable[P, R]:
    lock = RLock()

    def wrapper(*args: P.args, **kwargs: P.kwargs) -> R:
        key = cs.key_fn(*args, **kwargs)

        is_owner = False
        with lock:
            if (ret := cs.cache[key]) is not _empty:
                return ret

            # ... or it could be computed somewhere else, join there.
            f = cs.futures.get(key)
            if f:
                assert isinstance(f, cf.Future)
            else:
                cs.futures[key] = f = cf.Future[R]()
                is_owner = True

        # Release lock to allow function to run
        if not is_owner:
            with hide_frame:
                return _result(f)

        try:
            with hide_frame:
                ret = fn(*args, **kwargs)
        except BaseException as exc:
            exc = clone_exc(exc)  # Protect from mutation by outer frame
            f.set_exception(exc)
            with lock:
                cs.futures.pop(key)
            raise
        else:
            f.set_result(ret)
            with lock:
                cs.cache[key] = ret
                cs.futures.pop(key)
            return ret

    return wrapper


def _async_memoize[K: Hashable, **P, R](
    fn: Callable[P, Awaitable[R]],
    cs: _CacheState[K, R],
) -> Callable[P, Awaitable[R]]:
    async def wrapper(*args: P.args, **kwargs: P.kwargs) -> R:
        key = cs.key_fn(*args, **kwargs)

        if (ret := cs.cache[key]) is not _empty:
            return ret

        # ... or it could be computed somewhere else, join there.
        if f := cs.futures.get(key):
            assert isinstance(f, asyncio.Future)
            with hide_frame:
                return await f
        cs.futures[key] = f = asyncio.Future[R]()

        # NOTE: fn() is not within threading.Lock, thus it's not thread safe
        # NOTE: but it's async-safe because this `await` is only one here.
        try:
            with hide_frame:
                ret = await fn(*args, **kwargs)
        except BaseException as exc:
            exc = clone_exc(exc)
            f.set_exception(exc)
            cs.futures.pop(key)
            raise
        else:
            f.set_result(ret)
            cs.cache[key] = ret
            cs.futures.pop(key)
            return ret

    return wrapper


# ----------------------- wrapper with batching support ----------------------


class _BatchedQuery[K: Hashable, T, R]:
    def __init__(
        self, cs: _CacheState[K, R], *tokens: T, aio: bool = False
    ) -> None:
        self._cs = cs
        self._keys = [cs.key_fn(t) for t in tokens]  # All keys with duplicates

        self.jobs: list[tuple[K, Some[T] | None, AnyFuture[R]]] = []
        self._done: dict[K, R] = {}

        for k, t in dict(zip(self._keys, tokens)).items():
            # If this key is processing right now, wait till its done ...
            if f := cs.futures.get(k):  # ! Requires sync
                self.jobs.append((k, None, f))  # Wait for this

            # ... else check if it's done ...
            elif (r := cs.cache[k]) is not _empty:  # ! Requires sync
                self._done[k] = r

            # ... otherwise schedule a new job.
            else:
                f = asyncio.Future[R]() if aio else cf.Future[R]()
                self.jobs.append((k, Some(t), f))  # Resolve this manually
                cs.futures[k] = f  # ! Requires sync

    @property
    def pending_jobs(self) -> list[Job[T, R]]:
        return [(a.x, f) for _, a, f in self.jobs if a]

    def running_as[F: AnyFuture](self, tp: type[F]) -> set[F]:
        return {f for _, a, f in self.jobs if not a and isinstance(f, tp)}

    def sync(self, stash: Mapping[K, R]) -> None:
        for k, r in stash.items():
            self._done[k] = self._cs.cache[k] = r

        # Force next callers to use cache  # ! optional
        for k, _, _ in self.jobs:
            self._cs.futures.pop(k, None)

    def result(self) -> list[R]:
        return [self._done[k] for k in self._keys]


def _sync_memoize_batched[K: Hashable, T, R](
    fn: BatchFn[T, R], cs: _CacheState[K, R]
) -> BatchFn[T, R]:
    lock = RLock()

    def wrapper(tokens: Iterable[T]) -> list[R]:
        with lock:
            q = _BatchedQuery(cs, *tokens)

        stash: dict[K, R] = {}
        try:
            if jobs := q.pending_jobs:
                dispatch(fn, *jobs)

            if fs := q.running_as(cf.Future):
                cf.wait(fs)

            stash, err = gather_fs((k, f) for k, _, f in q.jobs)
        finally:
            if q.jobs:
                with lock:
                    q.sync(stash)

        if err is None:
            return q.result()
        with hide_frame:
            raise err

    return wrapper


def _async_memoize_batched[K: Hashable, T, R](
    fn: ABatchFn[T, R], cs: _CacheState[K, R]
) -> ABatchFn[T, R]:
    async def wrapper(tokens: Iterable[T]) -> list[R]:
        q = _BatchedQuery(cs, *tokens, aio=True)

        stash: dict[K, R] = {}
        try:
            if jobs := q.pending_jobs:
                await adispatch(fn, *jobs)

            if fs := q.running_as(asyncio.Future):
                await asyncio.wait(fs)

            stash, err = gather_fs((k, f) for k, _, f in q.jobs)
        finally:
            q.sync(stash)

        if err is None:
            return q.result()
        with hide_frame:
            raise err

    return wrapper


# ------------------------------- decorations --------------------------------


def _memoize[K: Hashable, **P, R](
    fn: Callable[P, R],
    *,
    cs: _CacheState[K, Any],
    batched: bool,
) -> Callable[P, R]:
    if batched and iscoroutinefunction(fn):
        w = cast(
            'Callable[P, R]',
            _async_memoize_batched(cast('ABatchFn', fn), cs=cs),
        )
    elif batched:
        w = cast(
            'Callable[P, R]',
            _sync_memoize_batched(cast('BatchFn', fn), cs=cs),
        )
    elif iscoroutinefunction(fn):
        w = cast('Callable[P, R]', _async_memoize(fn, cs=cs))
    else:
        w = _sync_memoize(fn, cs=cs)

    w.running = cs.futures  # type: ignore[attr-defined]
    if isinstance(cs.cache, _WeakCache):
        w.wrefs = cs.cache.alive  # type: ignore[attr-defined]
    if isinstance(cs.cache, _StrongCache):
        w.cache = cs.cache.cache  # type: ignore[attr-defined]

    return functools.update_wrapper(w, fn)


# ----------------------------- factory wrappers -----------------------------


def memoize(
    count: SupportsInt | None = None,
    *,
    nbytes: SupportsInt | None = None,
    batched: bool = False,
    policy: CachePolicy = None,
    key_fn: KeyFn = make_key,
) -> Decorator:
    """Create caching decorator.

    Parameters:
    - count - max objects to store or None for unbound cache.
    - nbytes - max bytes to store.
    - policy - eviction policy, either "raw" (no eviction), or "lru"
      (evict oldest), or "mru" (evict most recent).
    - batched - set if callable supports batching.
    """
    count = -1 if count is None else int(count)
    nbytes = -1 if nbytes is None else si_bin(int(nbytes))
    if count >= 0 and nbytes >= 0:
        msg = 'Only one of `count`/`nbytes` can be used. Not both'
        raise ValueError(msg)

    # count/nbytes in -/- (unbound), -/0 or 0/- (off), -/+ (bytes), +/- (count)
    capacity = max(count, nbytes)
    if int(capacity) == 0:
        return functools.partial(  # type: ignore[return-value]
            _memoize,
            cs=_CacheState(_WeakCache(), key_fn),
            batched=batched,
        )

    if cache_cls := _CACHES.get(policy):
        make_node = _make_node
        # count/nbytes in -/- (unbound), -/+ (bytes), +/- (count)
        if capacity < 0:
            cache_cls = _Heap
        # count/nbytes in -/+ (bytes), +/- (count)
        elif nbytes > 0:
            make_node = _make_sized_node

        cache = cache_cls(capacity, make_node)
        return functools.partial(  # type: ignore[return-value]
            _memoize,
            cs=_CacheState(_StrongCache(cache=cache), key_fn),
            batched=batched,
        )

    msg = f'Unknown cache policy: "{policy}". Available: "{set(_CACHES)}"'
    raise ValueError(msg)


_CACHES: dict[CachePolicy, _CacheMaker] = {
    None: _Heap,
    'lru': _LruCache,
    'mru': _MruCache,
}
